/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: flow-expire.c,v 1.5 2001/02/10 04:40:11 maf Exp $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <strings.h>
#include "ftlib.h"
#include "report.h"
#include "support.h"

char *progname = "flow-expire";
int detach, debug;

void usage(void);

int main(int argc, char **argv)
{
  extern char *optarg;
  extern int debug;
  struct ftfile_ager fa;
  int i, doit;
  char wpname[MAXPATHLEN+1];

  bzero(&fa, sizeof fa);

  fa.max_files = 0;
  fa.max_bytes = 0;
  wpname[0] = '.'; wpname[1] = 0;

  while ((i = getopt(argc, argv, "w:d:e:E:h?")) != -1)

    switch (i) {

    case 'd': /* debug */
      debug = atoi(optarg);
      break;

    case 'e': /* expire */
      fa.max_files = atoi(optarg);
      break;

    case 'E': /* size of flow files */
      if ((fa.max_bytes = scan_size(optarg)) == -1) {
        fprintf(stderr, "error scanning size\n");
        exit (1);
      }
      break;

    case 'w': /* output pathname */
      if (strlen(optarg) > (MAXPATHLEN)) {
        fprintf(stderr, "Pathname too long\n");
        exit (1);
      }
      strcpy(wpname, optarg);
      break;

    case '?':
    case 'h':
      usage();
      exit (0);

    default:
      usage();
      exit (1);

    } /* switch */

  if (debug) {
    doit = 0;
    detach = 0;
    report_init(1);
  }
  else {
    doit = 1;
    detach = 1;
    report_init(0);
  }


  if (ftfile_scandir(&fa, wpname)) {
    report(LOG_ERR, "ftfile_scandir(): failed");
    exit (1);
  }
  if (debug)
    ftfile_dump(&fa);
  if (ftfile_expire(&fa, doit, (u_int32)0)) {
    report(LOG_ERR, "ftfile_expire(): failed");
    exit (1);
  }
  if (debug)
    ftfile_dump(&fa);

  return 0;

} /* main */

void usage(void) {

  fprintf(stderr, "flow expire:\n\n");
  fprintf(stderr, " -d #  debug level.\n");
  fprintf(stderr, " -e #  flow files to keep before expire. (0=disable)\n");
  fprintf(stderr, " -E #  size of flow files to keep before expire. (0=disable)\n");
  fprintf(stderr, "       use b,K,M,G as multipliers, ie 16M for 16*1024*1024\n"); 
  fprintf(stderr, " -w x  working directory pathname.\n");
  fprintf(stderr, "\n\n");

} /* usage */
