/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: flow-stat.c,v 1.11 2001/02/10 04:42:45 maf Exp $
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>
#include "ftlib.h"
#include "support.h"
#include "fmt.h"
#include "hash_ip.h"
#include "hash_ip2.h"
#include "sym.h"
#include "bytes.h"

/*
 * XXX TODO
 *
 * tally passed to formats
 *
 * cc passed to formats
 *
 * update all formats for v8 
 *
 */

int debug;

int tally; /* the n in print a tally every n lines */
char cc; /* comment character */
int ip_net_only;

int tbl_out1(int sort_order, int options, u_int *index, u_int64 *flows,
  u_int64 *octets, u_int64 *packets, u_int64 *duration, u_int nindex,
  u_int64 nflows, u_int64 noctets, u_int64 npackets, u_int64 nduration,
  char *title, struct sym_table sym[]);

struct jump {
    int (*where)(struct ftio *ftio, int sort_order, int options);
};

int format0(struct ftio *ftio, int sort_order, int options);
int format1(struct ftio *ftio, int sort_order, int options);
int format2(struct ftio *ftio, int sort_order, int options);
int format3(struct ftio *ftio, int sort_order, int options);
int format4(struct ftio *ftio, int sort_order, int options);
int format5(struct ftio *ftio, int sort_order, int options);
int format6(struct ftio *ftio, int sort_order, int options);
int format7(struct ftio *ftio, int sort_order, int options);
int format8(struct ftio *ftio, int sort_order, int options);
int format9(struct ftio *ftio, int sort_order, int options);
int format10(struct ftio *ftio, int sort_order, int options);
int format11(struct ftio *ftio, int sort_order, int options);
int format12(struct ftio *ftio, int sort_order, int options);
int format13(struct ftio *ftio, int sort_order, int options);
int format14(struct ftio *ftio, int sort_order, int options);
int format15(struct ftio *ftio, int sort_order, int options);
int format16(struct ftio *ftio, int sort_order, int options);
int format17(struct ftio *ftio, int sort_order, int options);
int format18(struct ftio *ftio, int sort_order, int options);
int format19(struct ftio *ftio, int sort_order, int options);
int format20(struct ftio *ftio, int sort_order, int options);

struct jump format[] = {{format0}, {format1}, {format2},
          {format3}, {format4}, {format5},
          {format6}, {format7}, {format8},
          {format9}, {format10}, {format11},
          {format12}, {format13}, {format14},
          {format15}, {format16}, {format17},
          {format18}, {format19}, {format20},};

#define FORMATS 21  /* # of types of output */

struct stat0 {

  u_int64 nflows;     /* total # flows */
  u_int64 noctets;    /* total # octets */
  u_int64 npackets;   /* total # packets */

  u_int64 time;       /* total time in 1/1000 of flows */

  double  aflowtime;  /* average time of flow */
  double  aps;        /* average packet size */
  double  afs;        /* average flow size */
  double  apf;        /* average packets per flow */
  double  fps;        /* average flows per second */
  double  aos;        /* average octets per second */

  u_int64 start;      /* earliest flow time */
  u_int64 end;        /* latest flow time */

  /* average packet sizes */
  u_int64 psize32;    /* bytes/packet 1    <= p <= 32 */
  u_int64 psize64;    /* bytes/packet 32   < p <= 64  */
  u_int64 psize96; u_int64 psize128; u_int64 psize160; u_int64 psize192;
  u_int64 psize224; u_int64 psize256; u_int64 psize288; u_int64 psize320;
  u_int64 psize352; u_int64 psize384; u_int64 psize416; u_int64 psize448;
  u_int64 psize480; u_int64 psize512; u_int64 psize544; u_int64 psize576;
  u_int64 psize1024; u_int64 psize1536; u_int64 psize2048; u_int64 psize2560;
  u_int64 psize3072; u_int64 psize3584; u_int64 psize4096; u_int64 psize4608;

  /* packets per flow */
  u_int64 fpsize1;    /* packets/flow = 1 */
  u_int64 fpsize2;    /* packets/flow = 2 */
  u_int64 fpsize4;    /* packets/flow 2 < p <= 4 */
  u_int64 fpsize8;    /* packets/flow 4 < p <= 8 */
  u_int64 fpsize12; u_int64 fpsize16; u_int64 fpsize20; u_int64 fpsize24;
  u_int64 fpsize28; u_int64 fpsize32; u_int64 fpsize36; u_int64 fpsize40;
  u_int64 fpsize44; u_int64 fpsize48; u_int64 fpsize52; u_int64 fpsize60;
  u_int64 fpsize100; u_int64 fpsize200; u_int64 fpsize300; u_int64 fpsize400;
  u_int64 fpsize500; u_int64 fpsize600; u_int64 fpsize700; u_int64 fpsize800;
  u_int64 fpsize900;
  u_int64 fpsize_other; /* packets/flow 200 < p */

  /* octets per flow */
  u_int64 fosize32;     /* octets/flow 1    <= p <= 32 */
  u_int64 fosize64;     /* octets/flow 32   < p <= 64 */
  u_int64 fosize128;    /* octets/flow 64   < p <= 128 */
  u_int64 fosize256;    /* octets/flow 128   < p <= 256 */
  u_int64 fosize512; u_int64 fosize1280; u_int64 fosize2048;
  u_int64 fosize2816; u_int64 fosize3584; u_int64 fosize4352;
  u_int64 fosize5120; u_int64 fosize5888; u_int64 fosize6656;
  u_int64 fosize7424; u_int64 fosize8192; u_int64 fosize8960;
  u_int64 fosize9728; u_int64 fosize10496; u_int64 fosize11264;
  u_int64 fosize12032; u_int64 fosize12800; u_int64 fosize13568;
  u_int64 fosize14336; u_int64 fosize15104; u_int64 fosize15872;
  u_int64 fosize_other; /* octets/flow 15872   < p */

  /* time per flow */
  u_int64 ftime10;    /* time/flow 1 <= p <= 10 */
  u_int64 ftime50;    /* time/flow 10 < p <= 50 */
  u_int64 ftime100; u_int64 ftime200; u_int64 ftime500; u_int64 ftime1000;
  u_int64 ftime2000; u_int64 ftime3000; u_int64 ftime4000; u_int64 ftime5000;
  u_int64 ftime6000; u_int64 ftime7000; u_int64 ftime8000; u_int64 ftime9000;
  u_int64 ftime10000; u_int64 ftime12000; u_int64 ftime14000;
  u_int64 ftime16000; u_int64 ftime18000; u_int64 ftime20000;
  u_int64 ftime22000; u_int64 ftime24000; u_int64 ftime26000;
  u_int64 ftime28000; u_int64 ftime30000;
  u_int64 ftime_other;  /* time/flow 2000 < p */

};

void usage();

int main(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  extern char cc;
  extern struct jump format[];
  int x, y, ret, i, format_index;
  int sort_order, print_header;
  int options;
  struct ftio ftio;
  struct ftprof ftp;

  /* profile */
  ftprof_start (&ftp);

  options = 0;
  format_index = 0;
  ip_net_only = 0;
  print_header = 0;
  cc = '#';
  sort_order = 0;

  while ((i = getopt(argc, argv, "c:d:f:h?nCpPs:S:t:")) != -1)
    switch (i) {

    case 'c': /* comment character */
      cc = optarg[0];
      break;

    case 'C': /* summarize ip to classful networks */
      ip_net_only = 1;
      break;

    case 'd': /* debug */
      debug = atoi(optarg);
      break;

    case 'f': /* format */
      format_index = atoi(optarg);
      break;

    case 'h': /* help */
    case '?': /* help */
      usage();
      exit (0);
      break;

    case 'n': /* use names */
      options |= FT_OPT_NAMES;
      break;

    case 'p': /* print header */
      print_header = 1;
      break;

    case 'P': /* percent's */
      options |= FT_OPT_PERCENT;
      break;

    case 's': /* sort low to high on field n */
      sort_order = (atoi(optarg)+1) * -1;
      break;

    case 'S': /* sort high to low on field n */
      sort_order = (atoi(optarg)+1);
      break;

    case 't': /* tallies */
      options |= FT_OPT_TALLY;
      tally = atoi(optarg);
      break;

    default:
      usage();
      exit (1);
      break;

    } /* switch */

  if (format_index >= FORMATS) {
    fprintf(stderr, "no such format, %d\n", format_index);
    exit (1);
  }

  /* read from stdin */
  if (ftio_init(&ftio, 0, FT_IO_FLAG_READ) < 0) {
    fprintf(stderr, "ftio_init(): failed\n");
    exit (1);
  }

  /* print cmdl args */
  printf("%c\n%c output from: ", cc, cc);
  for (x = 0; x < argc; ++x) {
    for (y = 0; y < strlen(argv[x]); ++y) {
      if (isprint((int)argv[x][y])) putc(argv[x][y], stdout);
    }
    putc (' ', stdout);
  }
  putc ('\n', stdout);

  if (print_header) {
    ftio_header_print(&ftio, stdout, cc);
  }

  ret = format[format_index].where(&ftio, sort_order, options);

  if ((!ret) && (debug > 0)) {
    ftprof_end(&ftp, ftio_get_rec_total(&ftio));
    ftprof_print(&ftp, argv[0], stderr);
  }

  ret = 0;

  return ret;

} /* main */

void usage() {

  fprintf(stderr, "flow-stat:\n\n");
  fprintf(stderr, " -dn  set debug level to n.\n");
  fprintf(stderr, " -fn  format n.\n");
  fprintf(stderr, " -n   use symbolic names for ports and protocols\n");
  fprintf(stderr, " -p   print flow header.\n");
  fprintf(stderr, " -P   use percents instead of counters\n");
  fprintf(stderr, " -C   summarize IP host addreses to classful net addresses.\n");
  fprintf(stderr, " -Sn  sort descending on field n.\n");
  fprintf(stderr, " -sn  sort ascending on field n.\n");
  fprintf(stderr, " -tn  tally totals every n output lines when used with -P.\n");
  fprintf(stderr, "\n formats:\n");
  fprintf(stderr, "    0 Summary\n");
  fprintf(stderr, "    1 Average packet size distribution histogram\n");
  fprintf(stderr, "    2 Packets per flow distribution histogram\n");
  fprintf(stderr, "    3 Octets per flow flow distribution histogram\n");
  fprintf(stderr, "    5 UDP/TCP destination port flows,octets,packets,duration histogram\n");
  fprintf(stderr, "    6 UDP/TCP source port flows,octets,packets,duration histogram\n");
  fprintf(stderr, "    7 UDP/TCP port flows,octets,packets,duration histogram\n");
  fprintf(stderr, "    8 Destination IP flows,octets,packets,duration\n");
  fprintf(stderr, "    9 Source IP flows,octets,packets,duration\n");
  fprintf(stderr, "   10 Source/Destination IP flows,octets,packets,duration\n");
  fprintf(stderr, "   11 IP flows,octets,packets,duration\n");
  fprintf(stderr, "   12 flows,octets,packets,duration by IP protocol\n");
  fprintf(stderr, "   13 octets for flow duration plot data\n");
  fprintf(stderr, "   14 packets for flow duration plot data\n");
  fprintf(stderr, "   15 short summary\n");
  fprintf(stderr, "   16 Next hop flows,octets,packets,duration\n");
  fprintf(stderr, "   17 Input interface flows,octets,packets,duration\n");
  fprintf(stderr, "   18 Output interface flows,octets,packets,duration\n");
  fprintf(stderr, "   19 Source AS flows,octets,packets,duration\n");
  fprintf(stderr, "   20 Destination AS flows,octets,packets,duration\n");
  fprintf(stderr, "   21 Source/Destination AS flows,octets,packets,duration\n");
  fprintf(stderr, "\n\n");

} /* usage */


/*
 * function: format0
 *
 *  prints long summary
 *
 * returns 0 for okay.
 */
int format0(struct ftio *ftio, int sort_order, int options)
{
  struct stat0 fs0;
  u_long p;
  char fmt_buf[256];
  struct fts2rec_gen *rec;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  bzero(&fs0, sizeof fs0);
  fs0.start = 0xFFFFFFFF;
  fs0.end = 0;

  while ((rec = ftio_read(ftio))) {

    fs0.nflows ++;
    fs0.noctets += rec->dOctets;
    fs0.npackets += rec->dPkts;

    if (rec->First < fs0.start)
      fs0.start = rec->First;

    if (rec->Last > fs0.end)
      fs0.end = rec->Last;

    p = rec->dOctets / rec->dPkts;

    if (p <= 32) ++ fs0.psize32;
    else if (p <= 64) ++ fs0.psize64; else if (p <= 96) ++ fs0.psize96;
    else if (p <= 128) ++ fs0.psize128; else if (p <= 160) ++ fs0.psize160;
    else if (p <= 192) ++ fs0.psize192; else if (p <= 224) ++ fs0.psize224;
    else if (p <= 256) ++ fs0.psize256; else if (p <= 288) ++ fs0.psize288;
    else if (p <= 320) ++ fs0.psize320; else if (p <= 352) ++ fs0.psize352;
    else if (p <= 384) ++ fs0.psize384; else if (p <= 416) ++ fs0.psize416;
    else if (p <= 448) ++ fs0.psize448; else if (p <= 480) ++ fs0.psize480;
    else if (p <= 512) ++ fs0.psize512; else if (p <= 544) ++ fs0.psize544;
    else if (p <= 576) ++ fs0.psize576; else if (p <= 1024) ++ fs0.psize1024;
    else if (p <= 1536) ++ fs0.psize1536; else if (p <= 2048) ++ fs0.psize2048;
    else if (p <= 2560) ++ fs0.psize2560; else if (p <= 3072) ++ fs0.psize3072;
    else if (p <= 3584) ++ fs0.psize3584; else if (p <= 4096) ++ fs0.psize4096;
    else if (p <= 4608) ++ fs0.psize4608;

    p = rec->dPkts;
    if (p <= 1) ++ fs0.fpsize1; else if (p <= 2) ++ fs0.fpsize2;
    else if (p <= 4) ++ fs0.fpsize4; else if (p <= 8) ++ fs0.fpsize8;
    else if (p <= 12) ++ fs0.fpsize12; else if (p <= 16) ++ fs0.fpsize16;
    else if (p <= 20) ++ fs0.fpsize20; else if (p <= 24) ++ fs0.fpsize24;
    else if (p <= 28) ++ fs0.fpsize28; else if (p <= 32) ++ fs0.fpsize32;
    else if (p <= 36) ++ fs0.fpsize36; else if (p <= 40) ++ fs0.fpsize40;
    else if (p <= 44) ++ fs0.fpsize44; else if (p <= 48) ++ fs0.fpsize48;
    else if (p <= 52) ++ fs0.fpsize52; else if (p <= 60) ++ fs0.fpsize60;
    else if (p <= 100) ++ fs0.fpsize100; else if (p <= 200) ++ fs0.fpsize200;
    else if (p <= 300) ++ fs0.fpsize300; else if (p <= 400) ++ fs0.fpsize400;
    else if (p <= 500) ++ fs0.fpsize500; else if (p <= 600) ++ fs0.fpsize600;
    else if (p <= 700) ++ fs0.fpsize700; else if (p <= 800) ++ fs0.fpsize800;
    else if (p <= 900) ++ fs0.fpsize900; else ++ fs0.fpsize_other;

    p = rec->dOctets;
    if (p <= 32) ++ fs0.fosize32;
    else if (p <= 64) ++ fs0.fosize64; else if (p <= 128) ++ fs0.fosize128;
    else if (p <= 256) ++ fs0.fosize256; else if (p <= 512) ++ fs0.fosize512;
    else if (p <= 1280) ++ fs0.fosize1280;
    else if (p <= 2048) ++ fs0.fosize2048;
    else if (p <= 2816) ++ fs0.fosize2816;
    else if (p <= 3584) ++ fs0.fosize3584;
    else if (p <= 4352) ++ fs0.fosize4352;
    else if (p <= 5120) ++ fs0.fosize5120;
    else if (p <= 5888) ++ fs0.fosize5888;
    else if (p <= 6656) ++ fs0.fosize6656;
    else if (p <= 7424) ++ fs0.fosize7424;
    else if (p <= 8192) ++ fs0.fosize8192;
    else if (p <= 8960) ++ fs0.fosize8960;
    else if (p <= 9728) ++ fs0.fosize9728;
    else if (p <= 10496) ++ fs0.fosize10496;
    else if (p <= 11264) ++ fs0.fosize11264;
    else if (p <= 12032) ++ fs0.fosize12032;
    else if (p <= 12800) ++ fs0.fosize12800;
    else if (p <= 13568) ++ fs0.fosize13568;
    else if (p <= 14336) ++ fs0.fosize14336;
    else if (p <= 15104) ++ fs0.fosize15104;
    else if (p <= 15872) ++ fs0.fosize15872;
    else ++ fs0.fosize_other;


    p = rec->Last - rec->First;
    fs0.time += p;

    if (p <= 10) ++ fs0.ftime10;
    else if (p <= 50) ++ fs0.ftime50; else if (p <= 100) ++ fs0.ftime100;
    else if (p <= 200) ++ fs0.ftime200; else if (p <= 500) ++ fs0.ftime500;
    else if (p <= 1000) ++ fs0.ftime1000; else if (p <= 2000) ++ fs0.ftime2000;
    else if (p <= 3000) ++ fs0.ftime3000; else if (p <= 4000) ++ fs0.ftime4000;
    else if (p <= 5000) ++ fs0.ftime5000; else if (p <= 6000) ++ fs0.ftime6000;
    else if (p <= 7000) ++ fs0.ftime7000; else if (p <= 8000) ++ fs0.ftime8000;
    else if (p <= 9000) ++ fs0.ftime9000;
    else if (p <= 10000) ++ fs0.ftime10000;
    else if (p <= 12000) ++ fs0.ftime12000;
    else if (p <= 14000) ++ fs0.ftime14000;
    else if (p <= 16000) ++ fs0.ftime16000;
    else if (p <= 18000) ++ fs0.ftime18000;
    else if (p <= 20000) ++ fs0.ftime20000;
    else if (p <= 22000) ++ fs0.ftime22000;
    else if (p <= 24000) ++ fs0.ftime24000;
    else if (p <= 26000) ++ fs0.ftime26000;
    else if (p <= 28000) ++ fs0.ftime28000;
    else if (p <= 30000) ++ fs0.ftime30000;
    else ++ fs0.ftime_other;

  }

  fs0.aflowtime = fs0.time / fs0.nflows;
  fs0.aps = fs0.noctets / fs0.npackets;
  fs0.afs = fs0.noctets / fs0.nflows;
  fs0.apf = fs0.npackets / fs0.nflows;
  fs0.fps = (float)fs0.nflows / ((fs0.end - fs0.start) / 1000);
  fs0.aos = ((float)(fs0.noctets*8) / 1000) / ((fs0.end - fs0.start) / 1000);

  strcpy(fmt_buf, "Total Flows                     : ");
  fmt_uint64(fmt_buf+34, fs0.nflows, FMT_JUST_LEFT);
  puts(fmt_buf);

  strcpy(fmt_buf, "Total Octets                    : ");
  fmt_uint64(fmt_buf+34, fs0.noctets, FMT_JUST_LEFT);
  puts(fmt_buf);

  strcpy(fmt_buf, "Total Packets                   : ");
  fmt_uint64(fmt_buf+34, fs0.npackets, FMT_JUST_LEFT);
  puts(fmt_buf);

  strcpy(fmt_buf, "Total Time (1/1000 secs)        : ");
  fmt_uint64(fmt_buf+34, fs0.time, FMT_JUST_LEFT);
  puts(fmt_buf);

  strcpy(fmt_buf, "Duration of data (1/1000 secs)  : ");
  fmt_uint64(fmt_buf+34, (fs0.end - fs0.start), FMT_JUST_LEFT);
  puts(fmt_buf);

  printf("Average flow time (1/1000 secs) : %4.4f\n", fs0.aflowtime);
  printf("Average packet size (octets)    : %4.4f\n", fs0.aps);
  printf("Average flow size (octets)      : %4.4f\n", fs0.afs);
  printf("Average packets per flow        : %4.4f\n", fs0.apf);
  printf("Average flows / second          : %4.4f\n", fs0.fps);
  printf("Average Kbits / second          : %4.4f\n", fs0.aos);

  printf("\n\n");

  printf("IP packet size distribution:\n");
  printf("   1-32   64   96  128  160  192  224  256  288  320  352  384  416  448  480\n   ");
  print_3float((float)fs0.psize32 / fs0.nflows);
  print_3float((float)fs0.psize64 / fs0.nflows);
  print_3float((float)fs0.psize96 / fs0.nflows);
  print_3float((float)fs0.psize128 / fs0.nflows);
  print_3float((float)fs0.psize160 / fs0.nflows);
  print_3float((float)fs0.psize192 / fs0.nflows);
  print_3float((float)fs0.psize224 / fs0.nflows);
  print_3float((float)fs0.psize256 / fs0.nflows);
  print_3float((float)fs0.psize288 / fs0.nflows);
  print_3float((float)fs0.psize320 / fs0.nflows);
  print_3float((float)fs0.psize352 / fs0.nflows);
  print_3float((float)fs0.psize384 / fs0.nflows);
  print_3float((float)fs0.psize416 / fs0.nflows);
  print_3float((float)fs0.psize448 / fs0.nflows);
  print_3float((float)fs0.psize480 / fs0.nflows);
  printf("\n\n");

  printf("    512  544  576 1024 1536 2048 2560 3072 3584 4096 4608\n   ");
  print_3float((float)fs0.psize512 / fs0.nflows);
  print_3float((float)fs0.psize544 / fs0.nflows);
  print_3float((float)fs0.psize576 / fs0.nflows);
  print_3float((float)fs0.psize1024 / fs0.nflows);
  print_3float((float)fs0.psize1536 / fs0.nflows);
  print_3float((float)fs0.psize2048 / fs0.nflows);
  print_3float((float)fs0.psize2560 / fs0.nflows);
  print_3float((float)fs0.psize3072 / fs0.nflows);
  print_3float((float)fs0.psize3584 / fs0.nflows);
  print_3float((float)fs0.psize4096 / fs0.nflows);
  print_3float((float)fs0.psize4608 / fs0.nflows);
  printf("\n\n");

  printf("Packets per flow distribution:\n");
  printf("      1    2    4    8   12   16   20   24   28   32   36   40   44   48   52\n   ");

  print_3float((float)fs0.fpsize1 / fs0.nflows);
  print_3float((float)fs0.fpsize2 / fs0.nflows);
  print_3float((float)fs0.fpsize4 / fs0.nflows);
  print_3float((float)fs0.fpsize8 / fs0.nflows);
  print_3float((float)fs0.fpsize12 / fs0.nflows);
  print_3float((float)fs0.fpsize16 / fs0.nflows);
  print_3float((float)fs0.fpsize20 / fs0.nflows);
  print_3float((float)fs0.fpsize24 / fs0.nflows);
  print_3float((float)fs0.fpsize28 / fs0.nflows);
  print_3float((float)fs0.fpsize32 / fs0.nflows);
  print_3float((float)fs0.fpsize36 / fs0.nflows);
  print_3float((float)fs0.fpsize40 / fs0.nflows);
  print_3float((float)fs0.fpsize44 / fs0.nflows);
  print_3float((float)fs0.fpsize48 / fs0.nflows);
  print_3float((float)fs0.fpsize52 / fs0.nflows);
  printf("\n\n     60  100  200  300  400  500  600  700  800  900 >900\n   ");
  print_3float((float)fs0.fpsize60 / fs0.nflows);
  print_3float((float)fs0.fpsize100 / fs0.nflows);
  print_3float((float)fs0.fpsize200 / fs0.nflows);
  print_3float((float)fs0.fpsize300 / fs0.nflows);
  print_3float((float)fs0.fpsize400 / fs0.nflows);
  print_3float((float)fs0.fpsize500 / fs0.nflows);
  print_3float((float)fs0.fpsize600 / fs0.nflows);
  print_3float((float)fs0.fpsize700 / fs0.nflows);
  print_3float((float)fs0.fpsize800 / fs0.nflows);
  print_3float((float)fs0.fpsize900 / fs0.nflows);
  print_3float((float)fs0.fpsize_other / fs0.nflows);
  printf("\n\n");
             
  printf("Octets per flow distribution:\n");
  printf("     32   64  128  256  512 1280 2048 2816 3584 4352 5120 5888 6656 7424 8192\n   ");
  print_3float((float)fs0.fosize32 / fs0.nflows);
  print_3float((float)fs0.fosize64 / fs0.nflows);
  print_3float((float)fs0.fosize128 / fs0.nflows);
  print_3float((float)fs0.fosize256 / fs0.nflows);
  print_3float((float)fs0.fosize512 / fs0.nflows);
  print_3float((float)fs0.fosize1280 / fs0.nflows);
  print_3float((float)fs0.fosize2048 / fs0.nflows);
  print_3float((float)fs0.fosize2816 / fs0.nflows);
  print_3float((float)fs0.fosize3584 / fs0.nflows);
  print_3float((float)fs0.fosize4352 / fs0.nflows);
  print_3float((float)fs0.fosize5120 / fs0.nflows);
  print_3float((float)fs0.fosize5888 / fs0.nflows);
  print_3float((float)fs0.fosize6656 / fs0.nflows);
  print_3float((float)fs0.fosize7424 / fs0.nflows);
  print_3float((float)fs0.fosize8192 / fs0.nflows);
  printf("\n\n   8960 9728 10496 11264 12032 12800 13568 14336 15104 15872 >15872\n   ");
  print_3float((float)fs0.fosize8960 / fs0.nflows);
  print_3float2((float)fs0.fosize9728 / fs0.nflows);
  print_3float2((float)fs0.fosize10496 / fs0.nflows);
  print_3float2((float)fs0.fosize11264 / fs0.nflows);
  print_3float2((float)fs0.fosize12032 / fs0.nflows);
  print_3float2((float)fs0.fosize12800 / fs0.nflows);
  print_3float2((float)fs0.fosize13568 / fs0.nflows);
  print_3float2((float)fs0.fosize14336 / fs0.nflows);
  print_3float2((float)fs0.fosize15104 / fs0.nflows);
  print_3float2((float)fs0.fosize15872 / fs0.nflows);
  print_3float2((float)fs0.fosize_other / fs0.nflows);
  printf("\n\n");

  printf("Flow time distribution:\n");
  printf("    10    50  100  200  500 1000 2000 3000 4000 5000 6000 7000 8000 9000 10000\n   ");
  print_3float((float)fs0.ftime10 / fs0.nflows);
  print_3float((float)fs0.ftime50 / fs0.nflows);
  print_3float((float)fs0.ftime100 / fs0.nflows);
  print_3float((float)fs0.ftime200 / fs0.nflows);
  print_3float((float)fs0.ftime500 / fs0.nflows);
  print_3float((float)fs0.ftime1000 / fs0.nflows);
  print_3float((float)fs0.ftime2000 / fs0.nflows);
  print_3float((float)fs0.ftime3000 / fs0.nflows);
  print_3float((float)fs0.ftime4000 / fs0.nflows);
  print_3float((float)fs0.ftime5000 / fs0.nflows);
  print_3float((float)fs0.ftime6000 / fs0.nflows);
  print_3float((float)fs0.ftime7000 / fs0.nflows);
  print_3float((float)fs0.ftime8000 / fs0.nflows);
  print_3float2((float)fs0.ftime9000 / fs0.nflows);
  print_3float2((float)fs0.ftime10000 / fs0.nflows);
  printf("\n\n  12000 14000 16000 18000 20000 22000 24000 26000 28000 30000 >30000\n   ");
  print_3float2((float)fs0.ftime12000 / fs0.nflows);
  print_3float2((float)fs0.ftime14000 / fs0.nflows);
  print_3float2((float)fs0.ftime16000 / fs0.nflows);
  print_3float2((float)fs0.ftime18000 / fs0.nflows);
  print_3float2((float)fs0.ftime20000 / fs0.nflows);
  print_3float2((float)fs0.ftime22000 / fs0.nflows);
  print_3float2((float)fs0.ftime24000 / fs0.nflows);
  print_3float2((float)fs0.ftime26000 / fs0.nflows);
  print_3float2((float)fs0.ftime28000 / fs0.nflows);
  print_3float2((float)fs0.ftime30000 / fs0.nflows);
  print_3float2((float)fs0.ftime_other / fs0.nflows);
  printf("\n\n");

  return 0;

} /* format0 */

/*
 * function: format1
 *
 *  Average packet size distribution histogram
 *
 * returns 0 for okay.
 */
int format1(struct ftio *ftio, int sort_order, int options)
{
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration, overflow, tmp;
  u_int index[65536];
  struct fts2rec_gen *rec;
  int32 i;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = overflow = 0;

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;

  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    tmp = rec->dOctets / rec->dPkts;

    /* shouldn't happen */
    if (tmp > 65535) {
      ++ overflow;
      continue;
    }

    flows[tmp] ++;
    octets[tmp] += rec->dOctets;
    packets[tmp] += rec->dPkts;
    duration[tmp] += (rec->Last - rec->First);
  }

  /* doesn't make sense here */
  options &= ~FT_OPT_NAMES;

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "pkt size  ", 
    (struct sym_table*)0L);

  return 0;
} /* format1 */


/*
 * function: format2
 *
 *  Packets per flow distribution histogram
 *
 * returns 0 for okay.
 */
int format2(struct ftio *ftio, int sort_order, int options)
{
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration, overflow;
  u_int index[65536];
  struct fts2rec_gen *rec;
  int32 i;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = overflow = 0;

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;


  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    if (rec->dPkts > 65535) {
      ++ overflow;
      continue;
    }

    flows[rec->dPkts] ++;
    octets[rec->dPkts] += rec->dOctets;
    packets[rec->dPkts] += rec->dPkts;
    duration[rec->dPkts] += (rec->Last - rec->First);
  }

  /* doesn't make sense here */
  options &= ~FT_OPT_NAMES;

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "packets   ",
    (struct sym_table*)0L);

  return 0;

} /* format2 */

/*
 * function: format3
 *
 *  Octets per flow flow distribution histogram
 *
 */
int format3(struct ftio *ftio, int sort_order, int options)
{
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration, overflow;
  u_int index[65536];
  struct fts2rec_gen *rec;
  int32 i;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = overflow = 0;

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;


  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    if (rec->dOctets > 65535) {
      ++ overflow;
      continue;
    }

    flows[rec->dOctets] ++;
    octets[rec->dOctets] += rec->dOctets;
    packets[rec->dOctets] += rec->dPkts;
    duration[rec->dOctets] += (rec->Last - rec->First);
  }

  /* doesn't make sense here */
  options &= ~FT_OPT_NAMES;


  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "octets    ",
    (struct sym_table*)0L);

  return 0;

} /* format3 */

/*
 * function: format4
 *
 *  ??
 *
 */
int format4(struct ftio *ftio, int sort_order, int options)
{

  printf("Not implemented.\n");
  return 0;
}

/*
 * function: format5
 *
 *  UDP/TCP destination port flows,octets,packets,duration histogram
 *
 * returns 0 for okay.
 */
int format5(struct ftio *ftio, int sort_order, int options)
{
  extern struct sym_table sym_tcp[];
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration;
  u_int index[65536];
  struct fts2rec_gen *rec_gen;
  struct fts2rec_v8_2 *rec_v8_2;
  struct ftver ftv;
  int32 i;

  ftio_get_ver(ftio, &ftv);

  if (((ftv.d_version != 5) && (ftv.d_version != 7) && (ftv.d_version != 6)) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 2)))) {
    fprintf(stderr, "Port information not present in export version\n");
    return -1;
  }

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;

  switch (ftv.d_version) {

    case 5:
    case 6:
    case 7:
      while ((rec_gen = ftio_read(ftio))) {
        /* ports only make sense for TCP and UDP */
        if ((rec_gen->prot != IPPROTO_UDP) && (rec_gen->prot != IPPROTO_TCP))
          continue;
    
        ++nflows;
        noctets += rec_gen->dOctets;
        npackets += rec_gen->dPkts;
        nduration += (rec_gen->Last - rec_gen->First);
    
        flows[rec_gen->dstport] ++;
        octets[rec_gen->dstport] += rec_gen->dOctets;
        packets[rec_gen->dstport] += rec_gen->dPkts;
        duration[rec_gen->dstport] += (rec_gen->Last - rec_gen->First);
      } /* while */
      break;

    case 8:
      switch (ftv.agg_version) {

        case 2:
          while ((rec_v8_2 = ftio_read(ftio))) {
            /* ports only make sense for TCP and UDP */
            if ((rec_v8_2->prot != IPPROTO_UDP) && (rec_v8_2->prot !=
              IPPROTO_TCP))
              continue;
      
            ++nflows;
            noctets += rec_v8_2->dOctets;
            npackets += rec_v8_2->dPkts;
            nduration += (rec_v8_2->Last - rec_v8_2->First);
      
            flows[rec_v8_2->dstport] += rec_v8_2->dFlows;
            octets[rec_v8_2->dstport] += rec_v8_2->dOctets;
            packets[rec_v8_2->dstport] += rec_v8_2->dPkts;
            duration[rec_v8_2->dstport] += (rec_v8_2->Last - rec_v8_2->First);
          } /* while */
          break;
      } /* switch ftv.agg_version */
      break;

  } /* switch ftv.d_version */

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "port      ", sym_tcp);

  return 0;
}

/*
 * function: format6
 *
 *  UDP/TCP source port flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format6(struct ftio *ftio, int sort_order, int options)
{
  extern struct sym_table sym_tcp[];
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration;
  u_int index[65536];
  struct fts2rec_gen *rec_gen;
  struct fts2rec_v8_2 *rec_v8_2;
  struct ftver ftv;
  int32 i;

  ftio_get_ver(ftio, &ftv);

  if (((ftv.d_version != 5) && (ftv.d_version != 7) && (ftv.d_version != 6)) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 2)))) {
    fprintf(stderr, "Port information not present in export version\n");
    return -1;
  }

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  switch (ftv.d_version) {

    case 5:
    case 6:
    case 7:
      while ((rec_gen = ftio_read(ftio))) {
        /* ports only make sense for TCP and UDP */
        if ((rec_gen->prot != IPPROTO_UDP) && (rec_gen->prot != IPPROTO_TCP))
          continue;
    
        ++nflows;
        noctets += rec_gen->dOctets;
        npackets += rec_gen->dPkts;
        nduration += (rec_gen->Last - rec_gen->First);
    
        flows[rec_gen->srcport] ++;
        octets[rec_gen->srcport] += rec_gen->dOctets;
        packets[rec_gen->srcport] += rec_gen->dPkts;
        duration[rec_gen->srcport] += (rec_gen->Last - rec_gen->First);
      } /* while */
      break;

    case 8:
      switch (ftv.agg_version) {

        case 2:
          while ((rec_v8_2 = ftio_read(ftio))) {
            /* ports only make sense for TCP and UDP */
            if ((rec_v8_2->prot != IPPROTO_UDP) && (rec_v8_2->prot !=
              IPPROTO_TCP))
              continue;
      
            ++nflows;
            noctets += rec_v8_2->dOctets;
            npackets += rec_v8_2->dPkts;
            nduration += (rec_v8_2->Last - rec_v8_2->First);
      
            flows[rec_v8_2->srcport] += rec_v8_2->dFlows;
            octets[rec_v8_2->srcport] += rec_v8_2->dOctets;
            packets[rec_v8_2->srcport] += rec_v8_2->dPkts;
            duration[rec_v8_2->srcport] += (rec_v8_2->Last - rec_v8_2->First);
          } /* while */
          break;
      } /* switch ftv.agg_version */
      break;

  } /* switch ftv.d_version */

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "port      ", sym_tcp);

  return 0;
} /* format6 */

/*
 * function: format7
 *
 *  UDP/TCP port flows,octets,packets,duration histogram
 *
 * returns 0 for okay.
 */
int format7(struct ftio *ftio, int sort_order, int options)
{
  extern struct sym_table sym_tcp[];
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration;
  u_int index[65536];
  struct fts2rec_gen *rec;
  struct ftver ftv;
  int32 i;

  ftio_get_ver(ftio, &ftv);

  if (((ftv.d_version != 5) && (ftv.d_version != 7) && (ftv.d_version != 6)) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 2)))) {
    fprintf(stderr, "Port information not present in export version\n");
    return -1;
  }

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;


  while ((rec = ftio_read(ftio))) {

    /* ports only make sense for TCP and UDP */
    if ((rec->prot != IPPROTO_UDP) && (rec->prot != IPPROTO_TCP))
      continue;

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    flows[rec->srcport] ++;
    octets[rec->srcport] += rec->dOctets;
    packets[rec->srcport] += rec->dPkts;
    duration[rec->srcport] += (rec->Last - rec->First);

/* XXX??? */

/*
    if (rec->srcport != rec->dstport) {
      flows[rec->dstport] ++;
      octets[rec->dstport] += rec->dOctets;
      packets[rec->dstport] += rec->dPkts;
      duration[rec->dstport] += (rec->Last - rec->First);
    }
*/
  }

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "port      ", sym_tcp);

  return 0;

} /* format7 */

/*
 * function: format8
 *
 *  Destination IP flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format8(struct ftio *ftio, int sort_order, int options)
{
  struct fts2rec_gen *rec;
  struct hash_ip_rec hrec;
  struct hash_ip_table ht;
  extern int ip_net_only;
  extern char cc;
  u_int64 nflows, noctets, npackets, nduration;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = noctets = npackets = nduration = 0;

  if (hash_ip_init(&ht)) {
    fprintf(stderr, "hash_ip_init() failed\n");
    return -1;
  }

  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    hrec.addr = rec->dstaddr;
    hrec.nflows = 1;
    hrec.noctets = rec->dOctets;
    hrec.npackets = rec->dPkts;
    hrec.etime = (rec->Last - rec->First);

    if (ip_net_only) {
      if ((hrec.addr & 0x80000000) == 0)
        hrec.addr &= 0xFF000000;
      else if ((hrec.addr & 0xC0000000) == 0x80000000)
        hrec.addr &= 0xFFFF0000;
      else if ((hrec.addr & 0xC0000000) == 0xC0000000)
        hrec.addr &= 0xFFFFFF00;
    }

    if (hash_ip_update(&ht, &hrec) == -1) {
      fprintf(stderr, "hash_ip_update() failed\n");
      hash_ip_free(&ht);
      return -1;
    }

  }

/*  hash_ip_stat_dump(&ht); */

  hash_ip_dump(&ht, cc, sort_order, options, nflows, noctets, npackets,
    nduration);

  if (hash_ip_free(&ht)) {
    fprintf(stderr, "hash_ip_free() failed\n");
    return -1;
  }

  return 0;

} /* format8 */

/*
 * function: format9
 *
 *  Source IP flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format9(struct ftio *ftio, int sort_order, int options)
{
  struct fts2rec_gen *rec;
  struct hash_ip_rec hrec;
  struct hash_ip_table ht;
  extern int ip_net_only;
  extern char cc;
  u_int64 nflows, noctets, npackets, nduration;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = noctets = npackets = nduration = 0;

  if (hash_ip_init(&ht)) {
    fprintf(stderr, "hash_ip_init() failed\n");
    return -1;
  }

  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    hrec.addr = rec->srcaddr;
    hrec.nflows = 1;
    hrec.noctets = rec->dOctets;
    hrec.npackets = rec->dPkts;
    hrec.etime = (rec->Last - rec->First);

    if (ip_net_only) {
      if ((hrec.addr & 0x80000000) == 0)
        hrec.addr &= 0xFF000000;
      else if ((hrec.addr & 0xC0000000) == 0x80000000)
        hrec.addr &= 0xFFFF0000;
      else if ((hrec.addr & 0xC0000000) == 0xC0000000)
        hrec.addr &= 0xFFFFFF00;
    }

    if (hash_ip_update(&ht, &hrec) == -1) {
      fprintf(stderr, "hash_ip_update() failed\n");
      hash_ip_free(&ht);
      return -1;
    }

  }

/*  hash_ip_stat_dump(&ht); */

  hash_ip_dump(&ht, cc, sort_order, options, nflows, noctets, npackets,
    nduration);

  if (hash_ip_free(&ht)) {
    fprintf(stderr, "hash_ip_free() failed\n");
    return -1;
  }

  return 0;
}

/*
 * function: format10
 *
 *  Source/Destination IP flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format10(struct ftio *ftio, int sort_order, int options)
{
  struct fts2rec_gen *rec;
  struct hash_ip2_rec hrec;
  struct hash_ip2_table ht;
  extern int ip_net_only;
  extern char cc;
  u_int64 nflows, noctets, npackets, nduration;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = noctets = npackets = nduration = 0;
  
  if (hash_ip2_init(&ht)) {
    fprintf(stderr, "hash_ip2_init() failed\n");
    return -1;
  }

  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    hrec.src_addr = rec->srcaddr;
    hrec.dst_addr = rec->dstaddr;
    hrec.nflows = 1;
    hrec.noctets = rec->dOctets;
    hrec.npackets = rec->dPkts;
    hrec.etime = (rec->Last - rec->First);

    if (ip_net_only) {
      if ((hrec.src_addr & 0x80000000) == 0)
        hrec.src_addr &= 0xFF000000;
      else if ((hrec.src_addr & 0xC0000000) == 0x80000000)
        hrec.src_addr &= 0xFFFF0000;
      else if ((hrec.src_addr & 0xC0000000) == 0xC0000000)
        hrec.src_addr &= 0xFFFFFF00;

      if ((hrec.dst_addr & 0x80000000) == 0)
        hrec.dst_addr &= 0xFF000000;
      else if ((hrec.dst_addr & 0xC0000000) == 0x80000000)
        hrec.dst_addr &= 0xFFFF0000;
      else if ((hrec.dst_addr & 0xC0000000) == 0xC0000000)
        hrec.dst_addr &= 0xFFFFFF00;
    }

    if (hash_ip2_update(&ht, &hrec) == -1) {
      fprintf(stderr, "hash_ip2_update() failed\n");
      hash_ip2_free(&ht);
      return -1;
    }

  }

/*  hash_ip2_stat_dump(&ht); */

  hash_ip2_dump(&ht, cc, sort_order, options, nflows, noctets, npackets,
    nduration);

  if (hash_ip2_free(&ht)) {
    fprintf(stderr, "hash_ip2_free() failed\n");
    return -1;
  }

  return 0;
}

/*
 * function: format11
 *
 *  IP flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format11(struct ftio *ftio, int sort_order, int options)
{
  struct fts2rec_gen *rec;
  struct hash_ip_rec hrec;
  struct hash_ip_table ht;
  extern int ip_net_only;
  extern char cc;
  u_int64 nflows, noctets, npackets, nduration;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = noctets = npackets = nduration = 0;

  if (hash_ip_init(&ht)) {
    fprintf(stderr, "hash_ip_init() failed\n");
    return -1;
  }

  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    hrec.addr = rec->srcaddr;
    hrec.nflows = 1;
    hrec.noctets = rec->dOctets;
    hrec.npackets = rec->dPkts;
    hrec.etime = (rec->Last - rec->First);

    if (ip_net_only) {
      if ((hrec.addr & 0x80000000) == 0)
        hrec.addr &= 0xFF000000;
      else if ((hrec.addr & 0xC0000000) == 0x80000000)
        hrec.addr &= 0xFFFF0000;
      else if ((hrec.addr & 0xC0000000) == 0xC0000000)
        hrec.addr &= 0xFFFFFF00;
    }

    if (hash_ip_update(&ht, &hrec) == -1) {
      fprintf(stderr, "hash_ip_update() failed\n");
      hash_ip_free(&ht);
      return -1;
    }

    hrec.addr = rec->dstaddr;
    hrec.nflows = 1;
    hrec.noctets = rec->dOctets;
    hrec.npackets = rec->dPkts;
    hrec.etime = (rec->Last - rec->First);

    if (ip_net_only) {
      if ((hrec.addr & 0x80000000) == 0)
        hrec.addr &= 0xFF000000;
      else if ((hrec.addr & 0xC0000000) == 0x80000000)
        hrec.addr &= 0xFFFF0000;
      else if ((hrec.addr & 0xC0000000) == 0xC0000000)
        hrec.addr &= 0xFFFFFF00;
    }

    if (hash_ip_update(&ht, &hrec) == -1) {
      fprintf(stderr, "hash_ip_update() failed\n");
      hash_ip_free(&ht);
      return -1;
    }

  }

/*  hash_ip_stat_dump(&ht); */

  hash_ip_dump(&ht, cc, sort_order, options, nflows, noctets, npackets,
    nduration);

  if (hash_ip_free(&ht)) {
    fprintf(stderr, "hash_ip_free() failed\n");
    return -1;
  }

  return 0;
}

/*
 * function: format12
 *
 *  flows,octets,packets,duration by IP protocol
 *
 */
int format12(struct ftio *ftio, int sort_order, int options)
{
  u_int64 flows[256], octets[256], packets[256], duration[256];
  u_int64 nflows, noctets, npackets, nduration;
  u_int index[256];
  struct fts2rec_gen *rec;
  int32 i;
  extern struct sym_table sym_ipprot[];

  if (ftio_check_generic(ftio) < 0)
    return -1;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  /* initialize index */
  for (i = 0; i < 256; ++i)
    index[i] = i;

  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    flows[rec->prot] ++;
    octets[rec->prot] += rec->dOctets;
    packets[rec->prot] += rec->dPkts;
    duration[rec->prot] += (rec->Last - rec->First);
  }

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    256, nflows, noctets, npackets, nduration, "protocol  ", sym_ipprot);

  return 0;
}

/*
 * function: format13
 *
 *  octets for flow duration plot data
 *
 */
int format13(struct ftio *ftio, int sort_order, int options)
{
  struct fts2rec_gen *rec;
  u_long ymin, ymax;
  u_long xmin, xmax;
  extern char cc;
  u_int64 nflows;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = 0;

  xmin = ymin = 0xFFFFFFFF;
  xmax = ymax = 0;

  printf("%c\n%c start      octets\n%c end        octets\n%c\n", cc, cc, cc, cc);

  while ((rec = ftio_read(ftio))) {

    ++nflows;

    if (rec->Last > xmax)
      xmax = rec->Last;

    if (rec->First < xmin)
      xmin = rec->First;

    if (rec->dOctets > ymax)
      ymax = rec->dOctets;

    if (rec->dOctets < ymin)
      ymin = rec->dOctets;

    printf("%-10lu   %-10lu\n", (u_long)rec->First,
      (u_long)rec->dOctets);
    printf("%-10lu   %-10lu\n\n", (u_long)rec->Last,
      (u_long)rec->dOctets);

  }

  printf("%c xmin=%-10lu   ymin=%-10lu   xmax=%-10lu   ymax=%-10lu\n",
    cc, xmin, ymin, xmax, ymax);

  return 0;
}

/*
 * function: format14
 *
 *  octets for flow duration plot data
 *
 */
int format14(struct ftio *ftio, int sort_order, int options)
{
  struct fts2rec_gen *rec;
  u_long ymin, ymax;
  u_long xmin, xmax;
  u_int64 nflows;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = 0;

  xmin = ymin = 0xFFFFFFFF;
  xmax = ymax = 0;

  printf("%c\n%c start     packets\n%c end       packets\n%c\n", cc, cc, cc, cc);

  while ((rec = ftio_read(ftio))) {

    if (rec->Last > xmax)
      xmax = rec->Last;

    if (rec->First < xmin)
      xmin = rec->First;

    if (rec->dOctets > ymax)
      ymax = rec->dPkts;

    if (rec->dOctets < ymin)
      ymin = rec->dPkts;

    printf("%-10lu    %-10lu\n", (u_long)rec->First,
      (u_long)rec->dPkts);
    printf("%-10lu    %-10lu\n\n", (u_long)rec->Last,
      (u_long)rec->dPkts);

  }

  printf("%c xmin=%-10lu   ymin=%-10lu   xmax=%-10lu   ymax=%-10lu\n",
    cc, xmin, ymin, xmax, ymax);

  return 0;
}


/*
 * function: format15
 *
 *  short summary
 *
 * returns 0 for okay.
 */
int format15(struct ftio *ftio, int sort_order, int options)
{
  u_int64 octets, packets;
  double mbytes;
  struct fts2rec_gen *rec;
  char fmt_buf[256];
  u_int len;
  extern char cc;
  u_int64 nflows;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = 0;

  octets = packets = 0LL;

  while ((rec = ftio_read(ftio))) {
    ++nflows;
    octets += rec->dOctets;
    packets += rec->dPkts;
  }

  mbytes = (double)octets / (double)(1000000);

  printf("%c\n%c octets            packets             mbytes\n%c\n", cc, cc, cc);

  len = fmt_uint64(fmt_buf, octets, FMT_PAD_RIGHT);
  len += fmt_uint64(fmt_buf+len, packets, FMT_PAD_RIGHT);

  printf("%s%3.3f\n", fmt_buf, mbytes);

  return 0;

} /* format15 */

/*
 * function: format16
 *
 *  Next hop flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format16(struct ftio *ftio, int sort_order, int options)
{
  struct fts2rec_gen *rec;
  struct hash_ip_rec hrec;
  struct hash_ip_table ht;
  extern int ip_net_only;
  extern char cc;
  u_int64 nflows, noctets, npackets, nduration;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  nflows = noctets = npackets = nduration = 0;

  if (hash_ip_init(&ht)) {
    fprintf(stderr, "hash_ip_init() failed\n");
    return -1;
  }

  while ((rec = ftio_read(ftio))) {

    ++nflows;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    hrec.addr = rec->nexthop;
    hrec.nflows = 1;
    hrec.noctets = rec->dOctets;
    hrec.npackets = rec->dPkts;
    hrec.etime = (rec->Last - rec->First);

    if (ip_net_only) {
      if ((hrec.addr & 0x80000000) == 0)
        hrec.addr &= 0xFF000000;
      else if ((hrec.addr & 0xC0000000) == 0x80000000)
        hrec.addr &= 0xFFFF0000;
      else if ((hrec.addr & 0xC0000000) == 0xC0000000)
        hrec.addr &= 0xFFFFFF00;
    }

    if (hash_ip_update(&ht, &hrec) == -1) {
      fprintf(stderr, "hash_ip_update() failed\n");
      hash_ip_free(&ht);
      return -1;
    }

  }

/*  hash_ip_stat_dump(&ht); */

  hash_ip_dump(&ht, cc, sort_order, options, nflows, noctets, npackets,
    nduration);

  if (hash_ip_free(&ht)) {
    fprintf(stderr, "hash_ip_free() failed\n");
    return -1;
  }

  return 0;
} /* format16 */

/*
 * function: format17
 *
 *  input interface flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format17(struct ftio *ftio, int sort_order, int options)
{
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration;
  u_int index[65536];
  struct fts2rec_gen *rec;
  int32 i;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;

  while ((rec = ftio_read(ftio))) {

    nflows ++;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    flows[rec->input] ++;
    octets[rec->input] += rec->dOctets;
    packets[rec->input] += rec->dPkts;
    duration[rec->input] += (rec->Last - rec->First);
  }

  /* doesn't make sense here */
  options &= ~FT_OPT_NAMES;

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "interface ",
    (struct sym_table*)0L);


  return 0;

} /* format17 */

/*
 * function: format18
 *
 *  output interface flows,octets,packets,duration 
 *
 * returns 0 for okay.
 */
int format18(struct ftio *ftio, int sort_order, int options)
{
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration;
  u_int   index[65536];
  struct fts2rec_gen *rec;
  int32 i;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;


  while ((rec = ftio_read(ftio))) {

    nflows++;
    noctets += rec->dOctets;
    npackets += rec->dPkts;
    nduration += (rec->Last - rec->First);

    flows[rec->output] ++;
    octets[rec->output] += rec->dOctets;
    packets[rec->output] += rec->dPkts;
    duration[rec->output] += (rec->Last - rec->First);
  }

  /* doesn't make sense here */
  options &= ~FT_OPT_NAMES;

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "interface ",
    (struct sym_table*)0L);


  return 0;

} /* format18 */

/*
 * function: format19
 *
 *  Source AS flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format19(struct ftio *ftio, int sort_order, int options)
{
  extern struct sym_table sym_asn[];
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration;
  u_int index[65536];
  struct fts2rec_v5 *rec_v5;
  struct fts2rec_v8_1 *rec_v8_1;
  struct fts2rec_v8_3 *rec_v8_3;
  struct fts2rec_v8_5 *rec_v8_5;
  struct ftver ftv;
  int32 i;

  ftio_get_ver(ftio, &ftv);

  if (((ftv.d_version != 5) && (ftv.d_version != 7) && (ftv.d_version != 6)) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 1))) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 3))) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 5)))) {
    fprintf(stderr, "AS information not present in export version\n");
    return -1;
  }

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  switch (ftv.d_version) {

    case 5:
    case 6:
    case 7:
      while ((rec_v5 = ftio_read(ftio))) {

        ++nflows;
        noctets += rec_v5->dOctets;
        npackets += rec_v5->dPkts;
        nduration += (rec_v5->Last - rec_v5->First);

        flows[rec_v5->src_as] ++;
        octets[rec_v5->src_as] += rec_v5->dOctets;
        packets[rec_v5->src_as] += rec_v5->dPkts;
        duration[rec_v5->src_as] += (rec_v5->Last - rec_v5->First);
      }
      break;

    case 8:
      switch (ftv.agg_method) {

        case 1:
          while ((rec_v8_1 = ftio_read(ftio))) {

            nflows ++;
            noctets += rec_v8_1->dOctets;
            npackets += rec_v8_1->dPkts;
            nduration += (rec_v8_1->Last - rec_v8_1->First);

            flows[rec_v8_1->src_as] += rec_v8_1->dFlows;
            octets[rec_v8_1->src_as] += rec_v8_1->dOctets;
            packets[rec_v8_1->src_as] += rec_v8_1->dPkts;
            duration[rec_v8_1->src_as] += (rec_v8_1->Last - rec_v8_1->First);
          }
          break;

        case 3:
          while ((rec_v8_3 = ftio_read(ftio))) {

            nflows ++;
            noctets += rec_v8_3->dOctets;
            npackets += rec_v8_3->dPkts;
            nduration += (rec_v8_3->Last - rec_v8_3->First);

            flows[rec_v8_3->src_as] += rec_v8_3->dFlows;
            octets[rec_v8_3->src_as] += rec_v8_3->dOctets;
            packets[rec_v8_3->src_as] += rec_v8_3->dPkts;
            duration[rec_v8_3->src_as] += (rec_v8_3->Last - rec_v8_3->First);
          }
          break;

        case 5:
          while ((rec_v8_5 = ftio_read(ftio))) {

            nflows ++;
            noctets += rec_v8_5->dOctets;
            npackets += rec_v8_5->dPkts;
            nduration += (rec_v8_5->Last - rec_v8_5->First);

            flows[rec_v8_5->src_as] += rec_v8_5->dFlows;
            octets[rec_v8_5->src_as] += rec_v8_5->dOctets;
            packets[rec_v8_5->src_as] += rec_v8_5->dPkts;
            duration[rec_v8_5->src_as] += (rec_v8_5->Last - rec_v8_5->First);
          }
          break;

      } /* switch ftv.agg_method */
      break;

  } /* switch ftv.d_version */

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "src AS    ", 
    sym_asn);


  return 0;
} /* format19 */

/*
 * function: format20
 *
 *  Source AS flows,octets,packets,duration
 *
 * returns 0 for okay.
 */
int format20(struct ftio *ftio, int sort_order, int options)
{
  extern struct sym_table sym_asn[];
  u_int64 flows[65536], octets[65536], packets[65536], duration[65536];
  u_int64 nflows, noctets, npackets, nduration;
  u_int index[65536];
  struct fts2rec_v5 *rec_v5;
  struct fts2rec_v8_1 *rec_v8_1;
  struct fts2rec_v8_4 *rec_v8_4;
  struct fts2rec_v8_5 *rec_v8_5;
  struct ftver ftv;
  int32 i;

  ftio_get_ver(ftio, &ftv);

  if (((ftv.d_version != 5) && (ftv.d_version != 7) && (ftv.d_version != 6)) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 1))) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 4))) &&
       (!((ftv.d_version == 8) && (ftv.agg_method == 5)))) {
    fprintf(stderr, "AS information not present in export version\n");
    return -1;
  }

  /* initialize index */
  for (i = 0; i < 65536; ++i)
    index[i] = i;

  bzero(&flows, sizeof flows);
  bzero(&octets, sizeof octets);
  bzero(&packets, sizeof packets);
  bzero(&duration, sizeof duration);
  nflows = noctets = npackets = nduration = 0;

  switch (ftv.d_version) {

    case 5:
    case 6:
    case 7:
      while ((rec_v5 = ftio_read(ftio))) {

        ++nflows;
        noctets += rec_v5->dOctets;
        npackets += rec_v5->dPkts;
        nduration += (rec_v5->Last - rec_v5->First);

        flows[rec_v5->dst_as] ++;
        octets[rec_v5->dst_as] += rec_v5->dOctets;
        packets[rec_v5->dst_as] += rec_v5->dPkts;
        duration[rec_v5->dst_as] += (rec_v5->Last - rec_v5->First);
      }
      break;

    case 8:
      switch (ftv.agg_method) {

        case 1:
          while ((rec_v8_1 = ftio_read(ftio))) {

            nflows ++;
            noctets += rec_v8_1->dOctets;
            npackets += rec_v8_1->dPkts;
            nduration += (rec_v8_1->Last - rec_v8_1->First);

            flows[rec_v8_1->dst_as] += rec_v8_1->dFlows;
            octets[rec_v8_1->dst_as] += rec_v8_1->dOctets;
            packets[rec_v8_1->dst_as] += rec_v8_1->dPkts;
            duration[rec_v8_1->dst_as] += (rec_v8_1->Last - rec_v8_1->First);
          }
          break;

        case 4:
          while ((rec_v8_4 = ftio_read(ftio))) {

            nflows ++;
            noctets += rec_v8_4->dOctets;
            npackets += rec_v8_4->dPkts;
            nduration += (rec_v8_4->Last - rec_v8_4->First);

            flows[rec_v8_4->dst_as] += rec_v8_4->dFlows;
            octets[rec_v8_4->dst_as] += rec_v8_4->dOctets;
            packets[rec_v8_4->dst_as] += rec_v8_4->dPkts;
            duration[rec_v8_4->dst_as] += (rec_v8_4->Last - rec_v8_4->First);
          }
          break;

        case 5:
          while ((rec_v8_5 = ftio_read(ftio))) {

            nflows ++;
            noctets += rec_v8_5->dOctets;
            npackets += rec_v8_5->dPkts;
            nduration += (rec_v8_5->Last - rec_v8_5->First);

            flows[rec_v8_5->dst_as] += rec_v8_5->dFlows;
            octets[rec_v8_5->dst_as] += rec_v8_5->dOctets;
            packets[rec_v8_5->dst_as] += rec_v8_5->dPkts;
            duration[rec_v8_5->dst_as] += (rec_v8_5->Last - rec_v8_5->First);
          }
          break;

      } /* switch ftv.agg_method */
      break;

  } /* switch ftv.d_version */

  tbl_out1(sort_order, options, index, flows, octets, packets, duration,
    65536, nflows, noctets, npackets, nduration, "dst AS    ", 
    sym_asn);

  return 0;

} /* format20 */

int tbl_out1(int sort_order, int options, u_int *index, u_int64 *flows,
  u_int64 *octets, u_int64 *packets, u_int64 *duration, u_int nindex,
  u_int64 nflows, u_int64 noctets, u_int64 npackets, u_int64 nduration,
  char *title, struct sym_table sym[])
{
  char fmt_buf[256];
  int32 i, start, end, increment, x;
  int s, len;
  extern char cc;
  u_int64 t_flows, t_octets, t_packets, t_duration;
  u_int plines;

  s = abs(sort_order);
  x = 0;
  t_flows = t_octets = t_packets = t_duration = 0LL;

  if (s == 0) /* no sorting */
    ;
  else if (s == 1) /* port */
    ;
  else if (s == 2) /* flows */
    sort_64uint64(flows, index, nindex);
  else if (s == 3) /* octets */
    sort_64uint64(octets, index, nindex);
  else if (s == 4) /* packets */
    sort_64uint64(packets, index, nindex);
  else if (s == 5) /* duration */
    sort_64uint64(duration, index, nindex);
  else {
    printf("%c can't sort on field %d.\n", cc, sort_order);
    return -1;
  }

  if (sort_order >= 0)
    start = nindex - 1, end = -1, increment = -1;
  else
    start = 0, end = nindex, increment = 1;

  /* header */
  if (options & FT_OPT_PERCENT)
    printf("%c\n%c %sflows    octets   packets  duration\n%c\n",
      cc, cc, title, cc);
  else
    printf("%c\n%c %sflows                 octets                packets               duration\n%c\n", cc, cc, title, cc);

  /*
   * if FT_OPT_PERCENT and we're doing tallying, then the # of lines
   * that will be output needs to be known first
   */
  for (i = start; i != end; i += increment)
    if (flows[index[i]])
      ++plines;

  for (i = start; i != end; i += increment) {


    if (flows[index[i]]) {

      /* print a tally line ? */
      if ( (options & FT_OPT_TALLY) && x && (!(x % tally))) {

        if (options & FT_OPT_PERCENT) {

          printf("%cSUB %-6.3f  %-6.3f   %-6.3f   %-6.3f   %-6.3f\n", 
            cc, ((double)x/(double)plines)*100,
            ((double)t_flows/(double)nflows)*100,
            ((double)t_octets/(double)noctets)*100,
            ((double)t_packets/(double)npackets)*100,
            ((double)t_duration/(double)nduration)*100);
        } /* options & FT_OPT_PERCENT */
      } /* print tally */

      if (options & FT_OPT_PERCENT) {

        printf("%-10.10s  %-6.3f   %-6.3f   %-6.3f   %-6.3f\n", 
          sym_pr_uint16(sym, (u_int16)index[i], FMT_JUST_LEFT,
          options), ((double)flows[index[i]]/(double)nflows)*100,
          ((double)octets[index[i]]/(double)noctets)*100,
          ((double)packets[index[i]]/(double)npackets)*100,
          ((double)duration[index[i]]/(double)nduration)*100);

      } else {

        len = fmt_uint64(fmt_buf, flows[index[i]], FMT_PAD_RIGHT);
        fmt_buf[len++] = ' '; fmt_buf[len++] = ' '; 

        len += fmt_uint64(fmt_buf+len, octets[index[i]], FMT_PAD_RIGHT);
        fmt_buf[len++] = ' '; fmt_buf[len++] = ' '; 

        len += fmt_uint64(fmt_buf+len, packets[index[i]],
          FMT_PAD_RIGHT);
        fmt_buf[len++] = ' '; fmt_buf[len++] = ' '; 

        len += fmt_uint64(fmt_buf+len, duration[index[i]],
          FMT_JUST_LEFT);

        printf("%-10.10s  %s\n", sym_pr_uint16(sym,
          (u_int16)index[i], FMT_JUST_LEFT, options), fmt_buf);
      }

      t_flows += flows[index[i]];
      t_octets += octets[index[i]];
      t_packets += packets[index[i]];
      t_duration += duration[index[i]];
      ++x;

    } /* flows[index[i]] */
  }

  return 0;

} /* tbl_out1 */


